/*
 * debug - debug_mem.c
 * 	version 0.0.1
 * 
 * Copyright (C) 2005, 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <debug_mem.h>
#include <errno.h>
#include <sys/mman.h>
#include <string.h>

#ifdef USE_IOMEM

#define IOMEM_DEV	"/dev/iomem"

static int iomem_fd;

int d_read(unsigned long addr, void *buff, size_t count)
{
	if (iomem_fd == -1)
		return -ENODEV;
	if (lseek(iomem_fd, addr, SEEK_SET) == -1)
		return -errno;
	if (read(iomem_fd, buff, count) < count)
		return -errno;
	return 0;
}
int d_write(unsigned long addr, const void *buff, size_t count)
{
	if (iomem_fd == -1)
		return -ENODEV;
	if (lseek(iomem_fd, addr, SEEK_SET) == -1)
		return -errno;
	if (write(iomem_fd, buff, count) < count)
		return -errno;
	return 0;
}

int d_meminit(void)
{
	iomem_fd = open(IOMEM_DEV, O_RDWR);
	if (iomem_fd == -1) {
		fprintf(stderr, "debug: Can't open iomem device(%s).\n"
				"Functionality is limited.\n", IOMEM_DEV);
		return -errno;
	}
	return 0;
}

#else

#define MEM_DEV		"/dev/mem"

static int mem_fd;
static unsigned long page_size;
static unsigned long page_mask;

#define PAGE_SIZE	(page_size)
#define PAGE_MASK	(page_mask)

int d_read(unsigned long addr, void *buff, size_t count)
{
	unsigned long mbeg, mlen;
	uint8_t *mapping;
	
	if (!count)
		return 0;
	
	if (mem_fd == -1)
		return -ENODEV;

	mbeg = addr & PAGE_MASK;
	mlen = (addr + count - mbeg + PAGE_SIZE - 1) & PAGE_MASK;

	mapping = mmap(NULL, mlen, PROT_READ, MAP_SHARED, mem_fd, mbeg);
	if (mapping == MAP_FAILED)
		return -errno;
	
	memcpy(buff, mapping + addr - mbeg, count);
	return munmap(mapping, mlen);
}

int d_write(unsigned long addr, const void *buff, size_t count)
{
	unsigned long mbeg, mlen;
	uint8_t *mapping;
	
	if (!count)
		return 0;
	
	if (mem_fd == -1)
		return -ENODEV;

	mbeg = addr & PAGE_MASK;
	mlen = (addr + count - mbeg + PAGE_SIZE - 1) & PAGE_MASK;

	mapping = mmap(NULL, mlen, PROT_WRITE, MAP_SHARED, mem_fd, mbeg);
	if (mapping == MAP_FAILED)
		return -errno;
	memcpy(mapping + addr - mbeg, buff, count);
	return munmap(mapping, mlen);
}

#define DEFINE_WRITE(__name, __type)					\
	int __name(unsigned long addr, const __type *buff)		\
	{								\
		unsigned long mbeg, mlen;				\
		char *mapping;						\
		__type *x;						\
									\
		if (mem_fd == -1)					\
			return -ENODEV;					\
									\
		mbeg = addr & PAGE_MASK;				\
		mlen = (addr + sizeof(__type) - mbeg + PAGE_SIZE - 1)	\
		       & PAGE_MASK;					\
									\
		mapping = mmap(NULL, mlen, PROT_WRITE, MAP_SHARED,	\
				mem_fd, mbeg);				\
		if (mapping == MAP_FAILED)				\
			return -errno;					\
		x = (__type *)(mapping + addr - mbeg);			\
		*x = *buff;						\
		return munmap(mapping, mlen);				\
	}

DEFINE_WRITE(d_writeb, uint8_t)
DEFINE_WRITE(d_writew, uint16_t)
DEFINE_WRITE(d_writel, uint32_t)

#define DEFINE_READ(__name, __type)					\
	int __name(unsigned long addr, __type *buff)			\
	{								\
		unsigned long mbeg, mlen;				\
		char *mapping;						\
		__type *x;						\
									\
		if (mem_fd == -1)					\
			return -ENODEV;					\
									\
		mbeg = addr & PAGE_MASK;				\
		mlen = (addr + sizeof(__type) - mbeg + PAGE_SIZE - 1)	\
		       & PAGE_MASK;					\
									\
		mapping = mmap(NULL, mlen, PROT_READ, MAP_SHARED, 	\
				mem_fd, mbeg);				\
		if (mapping == MAP_FAILED)				\
			return -errno;					\
									\
		x = (__type *)(mapping + addr - mbeg);			\
		*buff = *x;						\
		return munmap(mapping, mlen);				\
	}

DEFINE_READ(d_readb, uint8_t)
DEFINE_READ(d_readw, uint16_t)
DEFINE_READ(d_readl, uint32_t)

int d_meminit(void)
{
	mem_fd = open(MEM_DEV, O_RDWR);
	if (mem_fd == -1) {
		fprintf(stderr, "debug: Can't open mem device(%s).\n"
				"Functionality is limited.\n", MEM_DEV);
		return -errno;
	}
	page_size = getpagesize();
	page_mask = ~(page_size - 1);
	return 0;
}

#endif
